# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import os
import os.path
import mediafile
import time

from datetime import datetime
from pony.orm import db_session
from queue import Queue, Empty as QueueEmpty
from threading import Thread, Event

from .covers import find_cover_in_folder, CoverFile
from .db import Folder, Artist, Album, Track


logger = logging.getLogger(__name__)


class StatsDetails:
    def __init__(self):
        self.artists = 0
        self.albums = 0
        self.tracks = 0


class Stats:
    def __init__(self):
        self.scanned = 0
        self.added = StatsDetails()
        self.deleted = StatsDetails()
        self.errors = []


class ScanQueue(Queue):
    def _init(self, maxsize):
        self.queue = set()
        self.__last_got = None

    def _put(self, item):
        if self.__last_got != item:
            self.queue.add(item)

    def _get(self):
        self.__last_got = self.queue.pop()
        return self.__last_got


class Scanner(Thread):
    def __init__(
        self,
        force=False,
        extensions=None,
        follow_symlinks=False,
        progress=None,
        on_folder_start=None,
        on_folder_end=None,
        on_done=None,
    ):
        super().__init__()

        if extensions is not None and not isinstance(extensions, list):
            raise TypeError("Invalid extensions type")

        self.__force = force
        self.__extensions = extensions
        self.__follow_symlinks = follow_symlinks

        self.__progress = progress
        self.__on_folder_start = on_folder_start
        self.__on_folder_end = on_folder_end
        self.__on_done = on_done

        self.__stopped = Event()
        self.__queue = ScanQueue()
        self.__stats = Stats()

    scanned = property(lambda self: self.__stats.scanned)

    def __report_progress(self, folder_name, scanned):
        if self.__progress is None:
            return

        self.__progress(folder_name, scanned)

    def queue_folder(self, folder_name):
        if not isinstance(folder_name, str):
            raise TypeError("Expecting string, got " + str(type(folder_name)))

        self.__queue.put(folder_name)

    def run(self):
        while not self.__stopped.is_set():
            try:
                folder_name = self.__queue.get(False)
            except QueueEmpty:
                break

            with db_session:
                folder = Folder.get(name=folder_name, root=True)
                if folder is None:
                    continue

            self.__scan_folder(folder)

        self.prune()

        if self.__on_done is not None:
            self.__on_done()

    def stop(self):
        self.__stopped.set()

    def __scan_folder(self, folder):
        logger.info("Scanning folder %s", folder.name)

        if self.__on_folder_start is not None:
            self.__on_folder_start(folder)

        # Scan new/updated files
        to_scan = [folder.path]
        scanned = 0
        while not self.__stopped.is_set() and to_scan:
            path = to_scan.pop()
            for entry in os.scandir(path):
                if entry.name.startswith("."):
                    continue
                if entry.is_symlink() and not self.__follow_symlinks:
                    continue
                elif entry.is_dir():
                    to_scan.append(entry.path)
                elif entry.is_file() and self.__check_extension(entry.path):
                    self.scan_file(entry)
                    self.__stats.scanned += 1
                    scanned += 1

                    self.__report_progress(folder.name, scanned)

        # Remove files that have been deleted
        # Could be more efficient if done above
        if not self.__stopped.is_set():
            with db_session:
                for track in Track.select(lambda t: t.root_folder == folder):
                    if not os.path.exists(track.path) or not self.__check_extension(
                        track.path
                    ):
                        self.remove_file(track.path)

        # Remove deleted/moved folders and update cover art info
        folders = [folder]
        while not self.__stopped.is_set() and folders:
            f = folders.pop()

            with db_session:
                # f has been fetched from another session, refetch or Pony will complain
                f = Folder[f.id]

                if not f.root and not os.path.isdir(f.path):
                    f.delete()  # Pony will cascade
                    continue

                self.find_cover(f.path)
                folders += f.children

        if not self.__stopped.is_set():
            with db_session:
                Folder[folder.id].last_scan = int(time.time())

        if self.__on_folder_end is not None:
            self.__on_folder_end(folder)

    def prune(self):
        if self.__stopped.is_set():
            return

        with db_session:
            self.__stats.deleted.albums = Album.prune()
            self.__stats.deleted.artists = Artist.prune()
            Folder.prune()

    def __check_extension(self, path):
        if not self.__extensions:
            return True
        return os.path.splitext(path)[1][1:].lower() in self.__extensions

    @db_session
    def scan_file(self, path_or_direntry):
        if isinstance(path_or_direntry, str):
            path = path_or_direntry

            if not os.path.exists(path):
                return

            basename = os.path.basename(path)
            stat = os.stat(path)
        else:
            path = path_or_direntry.path
            basename = path_or_direntry.name
            stat = path_or_direntry.stat()

        try:
            path.encode("utf-8")  # Test for badly encoded paths
        except UnicodeError:
            self.__stats.errors.append(path)
            return

        mtime = int(stat.st_mtime)

        tr = Track.get(path=path)
        if tr is not None:
            if not self.__force and not mtime > tr.last_modification:
                return

            tag = self.__try_load_tag(path)
            if tag is None:
                self.remove_file(path)
                return
            trdict = {}
        else:
            tag = self.__try_load_tag(path)
            if tag is None:
                return

            trdict = {"path": path}

        artist = (self.__sanitize_str(tag.artist) or "[unknown]")[:255]
        album = (self.__sanitize_str(tag.album) or "[non-album tracks]")[:255]
        albumartist = (self.__sanitize_str(tag.albumartist) or artist)[:255]

        trdict["disc"] = tag.disc or 1
        trdict["number"] = tag.track or 1
        trdict["title"] = (self.__sanitize_str(tag.title) or basename)[:255]
        trdict["year"] = tag.year
        trdict["genre"] = tag.genre
        trdict["duration"] = int(tag.length)
        trdict["has_art"] = bool(tag.images)

        trdict["bitrate"] = tag.bitrate
        trdict["last_modification"] = mtime

        tralbum = self.__find_album(albumartist, album)
        trartist = self.__find_artist(artist)

        if tr is None:
            trdict["root_folder"] = self.__find_root_folder(path)
            trdict["folder"] = self.__find_folder(path)
            trdict["album"] = tralbum
            trdict["artist"] = trartist
            trdict["created"] = datetime.fromtimestamp(mtime)

            try:
                Track(**trdict)
                self.__stats.added.tracks += 1
            except ValueError:
                # Field validation error
                self.__stats.errors.append(path)
        else:
            if tr.album.id != tralbum.id:
                trdict["album"] = tralbum

            if tr.artist.id != trartist.id:
                trdict["artist"] = trartist

            try:
                tr.set(**trdict)
            except ValueError:
                # Field validation error
                self.__stats.errors.append(path)

    @db_session
    def remove_file(self, path):
        if not isinstance(path, str):
            raise TypeError("Expecting string, got " + str(type(path)))

        tr = Track.get(path=path)
        if not tr:
            return

        self.__stats.deleted.tracks += 1
        tr.delete()

    @db_session
    def move_file(self, src_path, dst_path):
        if not isinstance(src_path, str):
            raise TypeError("Expecting string, got " + str(type(src_path)))
        if not isinstance(dst_path, str):
            raise TypeError("Expecting string, got " + str(type(dst_path)))

        if src_path == dst_path:
            return

        tr = Track.get(path=src_path)
        if tr is None:
            return

        tr_dst = Track.get(path=dst_path)
        if tr_dst is not None:
            root = tr_dst.root_folder
            folder = tr_dst.folder
            self.remove_file(dst_path)
            tr.root_folder = root
            tr.folder = folder
        else:
            root = self.__find_root_folder(dst_path)
            folder = self.__find_folder(dst_path)
            tr.root_folder = root
            tr.folder = folder
        tr.path = dst_path

    @db_session
    def find_cover(self, dirpath):
        if not isinstance(dirpath, str):  # pragma: nocover
            raise TypeError("Expecting string, got " + str(type(dirpath)))

        if not os.path.exists(dirpath):
            return

        folder = Folder.get(path=dirpath)
        if folder is None:
            return

        album_name = None
        track = folder.tracks.select().first()
        if track is not None:
            album_name = track.album.name

        cover = find_cover_in_folder(folder.path, album_name)
        folder.cover_art = cover.name if cover is not None else None

    @db_session
    def add_cover(self, path):
        if not isinstance(path, str):  # pragma: nocover
            raise TypeError("Expecting string, got " + str(type(path)))

        folder = Folder.get(path=os.path.dirname(path))
        if folder is None:
            return

        cover_name = os.path.basename(path)
        if not folder.cover_art:
            folder.cover_art = cover_name
        else:
            album_name = None
            track = folder.tracks.select().first()
            if track is not None:
                album_name = track.album.name

            current_cover = CoverFile(folder.cover_art, album_name)
            new_cover = CoverFile(cover_name, album_name)
            if new_cover.score > current_cover.score:
                folder.cover_art = cover_name

    def __find_album(self, artist, album):
        ar = self.__find_artist(artist)
        al = ar.albums.select(lambda a: a.name == album).first()
        if al:
            return al

        al = Album(name=album, artist=ar)
        self.__stats.added.albums += 1

        return al

    def __find_artist(self, artist):
        ar = Artist.get(name=artist)
        if ar:
            return ar

        ar = Artist(name=artist)
        self.__stats.added.artists += 1

        return ar

    def __find_root_folder(self, path):
        path = os.path.dirname(path)
        for folder in Folder.select(lambda f: f.root):
            if path.startswith(folder.path):
                return folder

        raise Exception(
            "Couldn't find the root folder for '{}'.\nDon't scan files that aren't located in a defined music folder".format(
                path
            )
        )

    def __find_folder(self, path):
        children = []
        drive, _ = os.path.splitdrive(path)
        path = os.path.dirname(path)
        while path not in (drive, "/"):
            folder = Folder.get(path=path)
            if folder is not None:
                break

            created = datetime.fromtimestamp(os.path.getmtime(path))
            children.append(
                dict(
                    root=False, name=os.path.basename(path), path=path, created=created
                )
            )
            path = os.path.dirname(path)

        assert folder is not None
        while children:
            folder = Folder(parent=folder, **children.pop())

        return folder

    def __try_load_tag(self, path):
        try:
            return mediafile.MediaFile(path)
        except mediafile.UnreadableFileError:
            return None

    def __sanitize_str(self, value):
        if value is None:
            return None
        return value.replace("\x00", "").strip()

    def stats(self):
        return self.__stats



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import shlex
import time

from datetime import datetime, timedelta
from pony.orm import db_session, ObjectNotFound
from random import shuffle
from subprocess import Popen, DEVNULL
from threading import Thread, Event, RLock

from .db import Track

logger = logging.getLogger(__name__)


class Jukebox:
    def __init__(self, cmd):
        self.__cmd = shlex.split(cmd)
        self.__playlist = []
        self.__index = 0
        self.__offset = 0
        self.__start = None

        self.__thread = None
        self.__lock = RLock()
        self.__skip = Event()
        self.__stop = Event()

    playing = property(
        lambda self: self.__thread is not None and self.__thread.is_alive()
    )
    index = property(lambda self: self.__index)
    gain = property(lambda self: 1.0)
    playlist = property(lambda self: list(self.__playlist))

    @property
    def position(self):
        if self.__start is None:
            return 0
        return int((datetime.utcnow() - self.__start).total_seconds())

    def set(self, *tracks):
        self.clear()
        self.add(*tracks)

    def start(self):
        if self.playing or not self.__playlist:
            return

        self.__skip.clear()
        self.__stop.clear()
        self.__offset = 0
        self.__thread = Thread(target=self.__play_thread)
        self.__thread.start()

    def stop(self):
        if not self.playing:
            return

        self.__stop.set()

    def skip(self, index, offset):
        if index < 0 or index >= len(self.__playlist):
            raise IndexError()
        if offset < 0:
            raise ValueError()

        with self.__lock:
            self.__index = index
            self.__offset = offset
            self.__start = datetime.utcnow() - timedelta(seconds=offset)
        self.__skip.set()
        self.start()

    def add(self, *tracks):
        with self.__lock:
            with db_session:
                for t in tracks:
                    try:
                        self.__playlist.append(Track[t].path)
                    except ObjectNotFound:
                        pass

    def clear(self):
        with self.__lock:
            self.__playlist.clear()
            self.__index = 0
            self.__offset = 0

    def remove(self, index):
        try:
            with self.__lock:
                self.__playlist.pop(index)
                if index < self.__index:
                    self.__index -= 1
        except IndexError:
            pass

    def shuffle(self):
        with self.__lock:
            shuffle(self.__playlist)

    def setgain(self, gain):
        pass

    def terminate(self):
        self.__stop.set()
        if self.__thread is not None:
            self.__thread.join()

    def __play_thread(self):
        proc = None
        while not self.__stop.is_set():
            if self.__skip.is_set():
                proc.terminate()
                proc.wait()
                proc = None
                self.__skip.clear()

            if proc is None:
                with self.__lock:
                    proc = self.__play_file()
            elif proc.poll() is not None:
                with self.__lock:
                    self.__start = None
                    self.__index += 1
                    if self.__index >= len(self.__playlist):
                        break

                    proc = self.__play_file()

            time.sleep(0.1)

        proc.terminate()
        proc.wait()
        self.__start = None

    def __play_file(self):
        path = self.__playlist[self.__index]
        args = [
            a.replace("%path", path).replace("%offset", str(self.__offset))
            for a in self.__cmd
        ]

        self.__start = datetime.utcnow() - timedelta(seconds=self.__offset)
        self.__offset = 0

        logger.debug("Start playing with command %s", args)
        try:
            return Popen(args, stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL)
        except Exception:
            logger.exception("Failed to run play command")
            return None



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import importlib
import mimetypes
import os.path
import pkg_resources
import time

from datetime import datetime
from hashlib import sha1
from pony.orm import Database, Required, Optional, Set, PrimaryKey, LongStr
from pony.orm import ObjectNotFound, DatabaseError
from pony.orm import buffer
from pony.orm import min, avg, sum, count, exists
from pony.orm import db_session
from urllib.parse import urlparse, parse_qsl
from uuid import UUID, uuid4

SCHEMA_VERSION = "20200607"


def now():
    return datetime.now().replace(microsecond=0)


metadb = Database()


class Meta(metadb.Entity):
    _table_ = "meta"
    key = PrimaryKey(str, 32)
    value = Required(str, 256)


db = Database()


@db.on_connect(provider="sqlite")
def sqlite_case_insensitive_like(db, connection):
    cursor = connection.cursor()
    cursor.execute("PRAGMA case_sensitive_like = OFF")


class PathMixin:
    @classmethod
    def get(cls, *args, **kwargs):
        if kwargs:
            path = kwargs.pop("path", None)
            if path:
                kwargs["_path_hash"] = sha1(path.encode("utf-8")).digest()
        return db.Entity.get.__func__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        path = kwargs["path"]
        kwargs["_path_hash"] = sha1(path.encode("utf-8")).digest()
        db.Entity.__init__(self, *args, **kwargs)

    def __setattr__(self, attr, value):
        db.Entity.__setattr__(self, attr, value)
        if attr == "path":
            db.Entity.__setattr__(
                self, "_path_hash", sha1(value.encode("utf-8")).digest()
            )


class Folder(PathMixin, db.Entity):
    _table_ = "folder"

    id = PrimaryKey(int, auto=True)
    root = Required(bool, default=False)
    name = Required(str, autostrip=False)
    path = Required(str, 4096, autostrip=False)  # unique
    _path_hash = Required(buffer, column="path_hash")
    created = Required(datetime, precision=0, default=now)
    cover_art = Optional(str, nullable=True, autostrip=False)
    last_scan = Required(int, default=0)

    parent = Optional(lambda: Folder, reverse="children", column="parent_id")
    children = Set(lambda: Folder, reverse="parent")

    __alltracks = Set(
        lambda: Track, lazy=True, reverse="root_folder"
    )  # Never used, hide it. Could be huge, lazy load
    tracks = Set(lambda: Track, reverse="folder")

    stars = Set(lambda: StarredFolder)
    ratings = Set(lambda: RatingFolder)

    def as_subsonic_child(self, user):
        info = dict(
            id=str(self.id),
            isDir=True,
            title=self.name,
            album=self.name,
            created=self.created.isoformat(),
        )
        if not self.root:
            info["parent"] = str(self.parent.id)
            info["artist"] = self.parent.name
        if self.cover_art:
            info["coverArt"] = str(self.id)
        else:
            for track in self.tracks:
                if track.has_art:
                    info["coverArt"] = str(track.id)
                    break

        try:
            starred = StarredFolder[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except ObjectNotFound:
            pass

        try:
            rating = RatingFolder[user.id, self.id]
            info["userRating"] = rating.rating
        except ObjectNotFound:
            pass

        avgRating = avg(self.ratings.rating)
        if avgRating:
            info["averageRating"] = avgRating

        return info

    def as_subsonic_artist(self, user):  # "Artist" type in XSD
        info = dict(id=str(self.id), name=self.name)

        try:
            starred = StarredFolder[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except ObjectNotFound:
            pass

        return info

    def as_subsonic_directory(self, user, client):  # "Directory" type in XSD
        info = dict(
            id=str(self.id),
            name=self.name,
            child=[
                f.as_subsonic_child(user)
                for f in self.children.order_by(lambda c: c.name.lower())
            ]
            + [
                t.as_subsonic_child(user, client)
                for t in sorted(self.tracks, key=lambda t: t.sort_key())
            ],
        )
        if not self.root:
            info["parent"] = str(self.parent.id)

        return info

    @classmethod
    def prune(cls):
        query = cls.select(
            lambda self: not exists(t for t in Track if t.folder == self)
            and not exists(f for f in Folder if f.parent == self)
            and not self.root
        )
        total = 0
        while True:
            count = query.delete()
            total += count
            if not count:
                return total


class Artist(db.Entity):
    _table_ = "artist"

    id = PrimaryKey(UUID, default=uuid4)
    name = Required(str)  # unique
    albums = Set(lambda: Album)
    tracks = Set(lambda: Track)

    stars = Set(lambda: StarredArtist)

    def as_subsonic_artist(self, user):
        info = dict(
            id=str(self.id),
            name=self.name,
            # coverArt
            albumCount=self.albums.count(),
        )

        try:
            starred = StarredArtist[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except ObjectNotFound:
            pass

        return info

    @classmethod
    def prune(cls):
        return cls.select(
            lambda self: not exists(a for a in Album if a.artist == self)
            and not exists(t for t in Track if t.artist == self)
        ).delete()


class Album(db.Entity):
    _table_ = "album"

    id = PrimaryKey(UUID, default=uuid4)
    name = Required(str)
    artist = Required(Artist, column="artist_id")
    tracks = Set(lambda: Track)

    stars = Set(lambda: StarredAlbum)

    def as_subsonic_album(self, user):  # "AlbumID3" type in XSD
        info = dict(
            id=str(self.id),
            name=self.name,
            artist=self.artist.name,
            artistId=str(self.artist.id),
            songCount=self.tracks.count(),
            duration=sum(self.tracks.duration),
            created=min(self.tracks.created).isoformat(),
        )

        track_with_cover = self.tracks.select(
            lambda t: t.folder.cover_art is not None
        ).first()
        if track_with_cover is not None:
            info["coverArt"] = str(track_with_cover.folder.id)
        else:
            track_with_cover = self.tracks.select(lambda t: t.has_art).first()
            if track_with_cover is not None:
                info["coverArt"] = str(track_with_cover.id)

        if count(self.tracks.year) > 0:
            info["year"] = min(self.tracks.year)

        genre = ", ".join(self.tracks.genre.distinct())
        if genre:
            info["genre"] = genre

        try:
            starred = StarredAlbum[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except ObjectNotFound:
            pass

        return info

    def sort_key(self):
        year = min(map(lambda t: t.year if t.year else 9999, self.tracks))
        return "%i%s" % (year, self.name.lower())

    @classmethod
    def prune(cls):
        return cls.select(
            lambda self: not exists(t for t in Track if t.album == self)
        ).delete()


class Track(PathMixin, db.Entity):
    _table_ = "track"

    id = PrimaryKey(UUID, default=uuid4)
    disc = Required(int)
    number = Required(int)
    title = Required(str)
    year = Optional(int)
    genre = Optional(str, nullable=True)
    duration = Required(int)
    has_art = Required(bool, default=False)

    album = Required(Album, column="album_id")
    artist = Required(Artist, column="artist_id")

    bitrate = Required(int)

    path = Required(str, 4096, autostrip=False)  # unique
    _path_hash = Required(buffer, column="path_hash")
    created = Required(datetime, precision=0, default=now)
    last_modification = Required(int)

    play_count = Required(int, default=0)
    last_play = Optional(datetime, precision=0)

    root_folder = Required(Folder, column="root_folder_id")
    folder = Required(Folder, column="folder_id")

    __lastly_played_by = Set(lambda: User)  # Never used, hide it

    stars = Set(lambda: StarredTrack)
    ratings = Set(lambda: RatingTrack)

    def as_subsonic_child(self, user, prefs):
        info = dict(
            id=str(self.id),
            parent=str(self.folder.id),
            isDir=False,
            title=self.title,
            album=self.album.name,
            artist=self.artist.name,
            track=self.number,
            size=os.path.getsize(self.path) if os.path.isfile(self.path) else -1,
            contentType=self.mimetype,
            suffix=self.suffix(),
            duration=self.duration,
            bitRate=self.bitrate,
            path=self.path[len(self.root_folder.path) + 1 :],
            isVideo=False,
            discNumber=self.disc,
            created=self.created.isoformat(),
            albumId=str(self.album.id),
            artistId=str(self.artist.id),
            type="music",
        )

        if self.year:
            info["year"] = self.year
        if self.genre:
            info["genre"] = self.genre
        if self.has_art:
            info["coverArt"] = str(self.id)
        elif self.folder.cover_art:
            info["coverArt"] = str(self.folder.id)

        try:
            starred = StarredTrack[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except ObjectNotFound:
            pass

        try:
            rating = RatingTrack[user.id, self.id]
            info["userRating"] = rating.rating
        except ObjectNotFound:
            pass

        avgRating = avg(self.ratings.rating)
        if avgRating:
            info["averageRating"] = avgRating

        if (
            prefs is not None
            and prefs.format is not None
            and prefs.format != self.suffix()
        ):
            info["transcodedSuffix"] = prefs.format
            info["transcodedContentType"] = (
                mimetypes.guess_type("dummyname." + prefs.format, False)[0]
                or "application/octet-stream"
            )

        return info

    @property
    def mimetype(self):
        return mimetypes.guess_type(self.path, False)[0] or "application/octet-stream"

    def duration_str(self):
        ret = "%02i:%02i" % ((self.duration % 3600) / 60, self.duration % 60)
        if self.duration >= 3600:
            ret = "%02i:%s" % (self.duration / 3600, ret)
        return ret

    def suffix(self):
        return os.path.splitext(self.path)[1][1:].lower()

    def sort_key(self):
        return (
            self.album.artist.name
            + self.album.name
            + ("%02i" % self.disc)
            + ("%02i" % self.number)
            + self.title
        ).lower()


class User(db.Entity):
    _table_ = "user"

    id = PrimaryKey(UUID, default=uuid4)
    name = Required(str, 64)  # unique
    mail = Optional(str)
    password = Required(str, 40)
    salt = Required(str, 6)

    admin = Required(bool, default=False)
    jukebox = Required(bool, default=False)

    lastfm_session = Optional(str, 32, nullable=True)
    lastfm_status = Required(
        bool, default=True
    )  # True: ok/unlinked, False: invalid session

    last_play = Optional(Track, column="last_play_id")
    last_play_date = Optional(datetime, precision=0)

    clients = Set(lambda: ClientPrefs)
    playlists = Set(lambda: Playlist)
    __messages = Set(lambda: ChatMessage, lazy=True)  # Never used, hide it

    starred_folders = Set(lambda: StarredFolder, lazy=True)
    starred_artists = Set(lambda: StarredArtist, lazy=True)
    starred_albums = Set(lambda: StarredAlbum, lazy=True)
    starred_tracks = Set(lambda: StarredTrack, lazy=True)
    folder_ratings = Set(lambda: RatingFolder, lazy=True)
    track_ratings = Set(lambda: RatingTrack, lazy=True)

    def as_subsonic_user(self):
        return dict(
            username=self.name,
            email=self.mail,
            scrobblingEnabled=self.lastfm_session is not None and self.lastfm_status,
            adminRole=self.admin,
            settingsRole=True,
            downloadRole=True,
            uploadRole=False,
            playlistRole=True,
            coverArtRole=False,
            commentRole=False,
            podcastRole=False,
            streamRole=True,
            jukeboxRole=self.admin or self.jukebox,
            shareRole=False,
        )


class ClientPrefs(db.Entity):
    _table_ = "client_prefs"

    user = Required(User, column="user_id")
    client_name = Required(str, 32)
    PrimaryKey(user, client_name)
    format = Optional(str, 8, nullable=True)
    bitrate = Optional(int)


class StarredFolder(db.Entity):
    _table_ = "starred_folder"

    user = Required(User, column="user_id")
    starred = Required(Folder, column="starred_id")
    date = Required(datetime, precision=0, default=now)

    PrimaryKey(user, starred)


class StarredArtist(db.Entity):
    _table_ = "starred_artist"

    user = Required(User, column="user_id")
    starred = Required(Artist, column="starred_id")
    date = Required(datetime, precision=0, default=now)

    PrimaryKey(user, starred)


class StarredAlbum(db.Entity):
    _table_ = "starred_album"

    user = Required(User, column="user_id")
    starred = Required(Album, column="starred_id")
    date = Required(datetime, precision=0, default=now)

    PrimaryKey(user, starred)


class StarredTrack(db.Entity):
    _table_ = "starred_track"

    user = Required(User, column="user_id")
    starred = Required(Track, column="starred_id")
    date = Required(datetime, precision=0, default=now)

    PrimaryKey(user, starred)


class RatingFolder(db.Entity):
    _table_ = "rating_folder"
    user = Required(User, column="user_id")
    rated = Required(Folder, column="rated_id")
    rating = Required(int, min=1, max=5)

    PrimaryKey(user, rated)


class RatingTrack(db.Entity):
    _table_ = "rating_track"
    user = Required(User, column="user_id")
    rated = Required(Track, column="rated_id")
    rating = Required(int, min=1, max=5)

    PrimaryKey(user, rated)


class ChatMessage(db.Entity):
    _table_ = "chat_message"

    id = PrimaryKey(UUID, default=uuid4)
    user = Required(User, column="user_id")
    time = Required(int, default=lambda: int(time.time()))
    message = Required(str, 512)

    def responsize(self):
        return dict(
            username=self.user.name, time=self.time * 1000, message=self.message
        )


class Playlist(db.Entity):
    _table_ = "playlist"

    id = PrimaryKey(UUID, default=uuid4)
    user = Required(User, column="user_id")
    name = Required(str)
    comment = Optional(str)
    public = Required(bool, default=False)
    created = Required(datetime, precision=0, default=now)
    tracks = Optional(LongStr)

    def as_subsonic_playlist(self, user):
        tracks = self.get_tracks()
        info = dict(
            id=str(self.id),
            name=self.name
            if self.user.id == user.id
            else "[{}] {}".format(self.user.name, self.name),
            owner=self.user.name,
            public=self.public,
            songCount=len(tracks),
            duration=sum(map(lambda t: t.duration, tracks)),
            created=self.created.isoformat(),
        )
        if self.comment:
            info["comment"] = self.comment
        return info

    def get_tracks(self):
        if not self.tracks:
            return []

        tracks = []
        should_fix = False

        for t in self.tracks.split(","):
            try:
                tid = UUID(t)
                track = Track[tid]
                tracks.append(track)
            except (ValueError, ObjectNotFound):
                should_fix = True

        if should_fix:
            self.tracks = ",".join(map(lambda t: str(t.id), tracks))
            db.commit()

        return tracks

    def clear(self):
        self.tracks = ""

    def add(self, track):
        if isinstance(track, UUID):
            tid = track
        elif isinstance(track, Track):
            tid = track.id
        elif isinstance(track, str):
            tid = UUID(track)

        if self.tracks and len(self.tracks) > 0:
            self.tracks = "{},{}".format(self.tracks, tid)
        else:
            self.tracks = str(tid)

    def remove_at_indexes(self, indexes):
        tracks = self.tracks.split(",")
        for i in indexes:
            if i < 0 or i >= len(tracks):
                continue
            tracks[i] = None

        self.tracks = ",".join(t for t in tracks if t)


class RadioStation(db.Entity):
    _table_ = "radio_station"

    id = PrimaryKey(UUID, default=uuid4)
    stream_url = Required(str)
    name = Required(str)
    homepage_url = Optional(str, nullable=True)
    created = Required(datetime, precision=0, default=now)

    def as_subsonic_station(self):
        info = dict(
            id=str(self.id),
            streamUrl=self.stream_url,
            name=self.name,
            homePageUrl=self.homepage_url,
        )
        return info


def parse_uri(database_uri):
    if not isinstance(database_uri, str):
        raise TypeError("Expecting a string")

    uri = urlparse(database_uri)
    args = dict(parse_qsl(uri.query))
    if uri.port is not None:
        args["port"] = uri.port

    if uri.scheme == "sqlite":
        path = uri.path
        if not path:
            path = ":memory:"
        elif path[0] == "/":
            path = path[1:]

        return dict(provider="sqlite", filename=path, create_db=True, **args)
    elif uri.scheme in ("postgres", "postgresql"):
        return dict(
            provider="postgres",
            user=uri.username,
            password=uri.password,
            host=uri.hostname,
            dbname=uri.path[1:],
            **args
        )
    elif uri.scheme == "mysql":
        args.setdefault("charset", "utf8mb4")
        args.setdefault("binary_prefix", True)
        return dict(
            provider="mysql",
            user=uri.username,
            passwd=uri.password,
            host=uri.hostname,
            db=uri.path[1:],
            **args
        )
    return dict()


def execute_sql_resource_script(respath):
    sql = pkg_resources.resource_string(__package__, respath).decode("utf-8")
    for statement in sql.split(";"):
        statement = statement.strip()
        if statement and not statement.startswith("--"):
            metadb.execute(statement)


def init_database(database_uri):
    settings = parse_uri(database_uri)

    metadb.bind(**settings)
    metadb.generate_mapping(check_tables=False)

    # Check if we should create the tables
    try:
        metadb.check_tables()
    except DatabaseError:
        with db_session:
            execute_sql_resource_script("schema/" + settings["provider"] + ".sql")
            Meta(key="schema_version", value=SCHEMA_VERSION)

    # Check for schema changes
    with db_session:
        version = Meta["schema_version"]
        if version.value < SCHEMA_VERSION:
            migrations = sorted(
                pkg_resources.resource_listdir(
                    __package__, "schema/migration/" + settings["provider"]
                )
            )
            for migration in migrations:
                date, ext = os.path.splitext(migration)
                if date <= version.value:
                    continue
                if ext == ".sql":
                    execute_sql_resource_script(
                        "schema/migration/{}/{}".format(settings["provider"], migration)
                    )
                elif ext == ".py":
                    m = importlib.import_module(
                        ".schema.migration.{}.{}".format(settings["provider"], date),
                        __package__,
                    )
                    m.apply(settings.copy())
            version.value = SCHEMA_VERSION

    # Hack for in-memory SQLite databases (used in tests), otherwise 'db' and 'metadb' would be two distinct databases
    # and 'db' wouldn't have any table
    if settings["provider"] == "sqlite" and settings["filename"] == ":memory:":
        db.provider = metadb.provider
    else:
        metadb.disconnect()
        db.bind(**settings)
        # Force requests to Meta to use the same connection as other tables
        metadb.provider = db.provider

    db.generate_mapping(check_tables=False)


def release_database():
    metadb.disconnect()
    db.disconnect()
    db.provider = metadb.provider = None
    db.schema = metadb.schema = None



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from flask import request

from ..db import Playlist, User, Track

from . import get_entity, api_routing
from .exceptions import Forbidden, MissingParameter, NotFound


@api_routing("/getPlaylists")
def list_playlists():
    query = Playlist.select(
        lambda p: p.user.id == request.user.id or p.public
    ).order_by(Playlist.name)

    username = request.values.get("username")
    if username:
        if not request.user.admin:
            raise Forbidden()

        user = User.get(name=username)
        if user is None:
            raise NotFound("User")

        query = Playlist.select(lambda p: p.user.name == username).order_by(
            Playlist.name
        )

    return request.formatter(
        "playlists",
        dict(playlist=[p.as_subsonic_playlist(request.user) for p in query]),
    )


@api_routing("/getPlaylist")
def show_playlist():
    res = get_entity(Playlist)
    if res.user.id != request.user.id and not res.public and not request.user.admin:
        raise Forbidden()

    info = res.as_subsonic_playlist(request.user)
    info["entry"] = [
        t.as_subsonic_child(request.user, request.client) for t in res.get_tracks()
    ]
    return request.formatter("playlist", info)


@api_routing("/createPlaylist")
def create_playlist():
    playlist_id, name = map(request.values.get, ["playlistId", "name"])
    # songId actually doesn't seem to be required
    songs = request.values.getlist("songId")
    playlist_id = uuid.UUID(playlist_id) if playlist_id else None

    if playlist_id:
        playlist = Playlist[playlist_id]

        if playlist.user.id != request.user.id and not request.user.admin:
            raise Forbidden()

        playlist.clear()
        if name:
            playlist.name = name
    elif name:
        playlist = Playlist(user=request.user, name=name)
    else:
        raise MissingParameter("playlistId or name")

    for sid in songs:
        sid = uuid.UUID(sid)
        track = Track[sid]
        playlist.add(track)

    return request.formatter.empty


@api_routing("/deletePlaylist")
def delete_playlist():
    res = get_entity(Playlist)
    if res.user.id != request.user.id and not request.user.admin:
        raise Forbidden()

    res.delete()
    return request.formatter.empty


@api_routing("/updatePlaylist")
def update_playlist():
    res = get_entity(Playlist, "playlistId")
    if res.user.id != request.user.id and not request.user.admin:
        raise Forbidden()

    playlist = res
    name, comment, public = map(request.values.get, ["name", "comment", "public"])
    to_add, to_remove = map(
        request.values.getlist, ["songIdToAdd", "songIndexToRemove"]
    )

    if name:
        playlist.name = name
    if comment:
        playlist.comment = comment
    if public:
        playlist.public = public in (True, "True", "true", 1, "1")

    to_add = map(uuid.UUID, to_add)
    to_remove = map(int, to_remove)

    for sid in to_add:
        track = Track[sid]
        playlist.add(track)

    playlist.remove_at_indexes(to_remove)

    return request.formatter.empty



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2019 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path

from pony.orm import select
from pony.orm import ObjectNotFound

from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Folder, Track, Artist, Album, User, RatingTrack, StarredTrack


class FolderManager:
    @staticmethod
    def get(id):
        try:
            id = int(id)
        except ValueError:
            raise ValueError("Invalid folder id")

        return Folder[id]

    @staticmethod
    def add(name, path):
        if Folder.get(name=name, root=True) is not None:
            raise ValueError("Folder '{}' exists".format(name))

        path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isdir(path):
            raise ValueError("The path doesn't exits or isn't a directory")
        if Folder.get(path=path) is not None:
            raise ValueError("This path is already registered")
        if any(path.startswith(p) for p in select(f.path for f in Folder if f.root)):
            raise ValueError("This path is already registered")
        if Folder.exists(lambda f: f.path.startswith(path)):
            raise ValueError("This path contains a folder that is already registered")

        folder = Folder(root=True, name=name, path=path)
        try:
            DaemonClient().add_watched_folder(path)
        except DaemonUnavailableError:
            pass

        return folder

    @staticmethod
    def delete(id):
        folder = FolderManager.get(id)
        if not folder.root:
            raise ObjectNotFound(Folder)

        try:
            DaemonClient().remove_watched_folder(folder.path)
        except DaemonUnavailableError:
            pass

        for user in User.select(lambda u: u.last_play.root_folder == folder):
            user.last_play = None
        RatingTrack.select(lambda r: r.rated.root_folder == folder).delete(bulk=True)
        StarredTrack.select(lambda s: s.starred.root_folder == folder).delete(bulk=True)

        Track.select(lambda t: t.root_folder == folder).delete(bulk=True)
        Album.prune()
        Artist.prune()
        Folder.select(lambda f: not f.root and f.path.startswith(folder.path)).delete(
            bulk=True
        )

        folder.delete()

    @staticmethod
    def delete_by_name(name):
        folder = Folder.get(name=name, root=True)
        if not folder:
            raise ObjectNotFound(Folder)
        FolderManager.delete(folder.id)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from . import api
from .exceptions import GenericError

methods = (
    "getVideos",
    "getAvatar",
    "getShares",
    "createShare",
    "updateShare",
    "deleteShare",
)


def unsupported():
    return GenericError("Not supported by Supysonic")


for m in methods:
    api.add_url_rule(
        "/{}".format(m), "unsupported", unsupported, methods=["GET", "POST"]
    )
    api.add_url_rule(
        "/{}.view".format(m), "unsupported", unsupported, methods=["GET", "POST"]
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from collections import OrderedDict
from datetime import datetime
from flask import request
from pony.orm import select

from ..db import Folder, Track, Artist, Album

from . import api_routing
from .exceptions import MissingParameter


@api_routing("/search")
def old_search():
    artist, album, title, anyf, count, offset, newer_than = map(
        request.values.get,
        ["artist", "album", "title", "any", "count", "offset", "newerThan"],
    )

    count = int(count) if count else 20
    offset = int(offset) if offset else 0
    newer_than = int(newer_than) / 1000 if newer_than else 0
    min_date = datetime.fromtimestamp(newer_than)

    if artist:
        query = select(
            t.folder.parent
            for t in Track
            if artist in t.folder.parent.name and t.folder.parent.created > min_date
        )
    elif album:
        query = select(
            t.folder
            for t in Track
            if album in t.folder.name and t.folder.created > min_date
        )
    elif title:
        query = Track.select(lambda t: title in t.title and t.created > min_date)
    elif anyf:
        folders = Folder.select(lambda f: anyf in f.name and f.created > min_date)
        tracks = Track.select(lambda t: anyf in t.title and t.created > min_date)
        res = folders[offset : offset + count]
        fcount = folders.count()
        if offset + count > fcount:
            toff = max(0, offset - fcount)
            tend = offset + count - fcount
            res = res[:] + tracks[toff:tend][:]

        return request.formatter(
            "searchResult",
            dict(
                totalHits=folders.count() + tracks.count(),
                offset=offset,
                match=[
                    r.as_subsonic_child(request.user)
                    if isinstance(r, Folder)
                    else r.as_subsonic_child(request.user, request.client)
                    for r in res
                ],
            ),
        )
    else:
        raise MissingParameter("search")

    return request.formatter(
        "searchResult",
        dict(
            totalHits=query.count(),
            offset=offset,
            match=[
                r.as_subsonic_child(request.user)
                if isinstance(r, Folder)
                else r.as_subsonic_child(request.user, request.client)
                for r in query[offset : offset + count]
            ],
        ),
    )


@api_routing("/search2")
def new_search():
    query = request.values["query"]
    (
        artist_count,
        artist_offset,
        album_count,
        album_offset,
        song_count,
        song_offset,
    ) = map(
        request.values.get,
        [
            "artistCount",
            "artistOffset",
            "albumCount",
            "albumOffset",
            "songCount",
            "songOffset",
        ],
    )

    artist_count = int(artist_count) if artist_count else 20
    artist_offset = int(artist_offset) if artist_offset else 0
    album_count = int(album_count) if album_count else 20
    album_offset = int(album_offset) if album_offset else 0
    song_count = int(song_count) if song_count else 20
    song_offset = int(song_offset) if song_offset else 0

    artists = select(
        t.folder.parent for t in Track if query in t.folder.parent.name
    ).limit(artist_count, artist_offset)
    albums = select(t.folder for t in Track if query in t.folder.name).limit(
        album_count, album_offset
    )
    songs = Track.select(lambda t: query in t.title).limit(song_count, song_offset)

    return request.formatter(
        "searchResult2",
        OrderedDict(
            (
                ("artist", [a.as_subsonic_artist(request.user) for a in artists]),
                ("album", [f.as_subsonic_child(request.user) for f in albums]),
                (
                    "song",
                    [t.as_subsonic_child(request.user, request.client) for t in songs],
                ),
            )
        ),
    )


@api_routing("/search3")
def search_id3():
    query = request.values["query"]
    (
        artist_count,
        artist_offset,
        album_count,
        album_offset,
        song_count,
        song_offset,
    ) = map(
        request.values.get,
        [
            "artistCount",
            "artistOffset",
            "albumCount",
            "albumOffset",
            "songCount",
            "songOffset",
        ],
    )

    artist_count = int(artist_count) if artist_count else 20
    artist_offset = int(artist_offset) if artist_offset else 0
    album_count = int(album_count) if album_count else 20
    album_offset = int(album_offset) if album_offset else 0
    song_count = int(song_count) if song_count else 20
    song_offset = int(song_offset) if song_offset else 0

    artists = Artist.select(lambda a: query in a.name).limit(
        artist_count, artist_offset
    )
    albums = Album.select(lambda a: query in a.name).limit(album_count, album_offset)
    songs = Track.select(lambda t: query in t.title).limit(song_count, song_offset)

    return request.formatter(
        "searchResult3",
        OrderedDict(
            (
                ("artist", [a.as_subsonic_artist(request.user) for a in artists]),
                ("album", [a.as_subsonic_album(request.user) for a in albums]),
                (
                    "song",
                    [t.as_subsonic_child(request.user, request.client) for t in songs],
                ),
            )
        ),
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2019 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import (
    current_app,
    flash,
    redirect,
    request,
    render_template,
    session,
    url_for,
)
from flask import Blueprint
from functools import wraps
from pony.orm import ObjectNotFound

from ..daemon.client import DaemonClient
from ..daemon.exceptions import DaemonUnavailableError
from ..db import Artist, Album, Track
from ..managers.user import UserManager

frontend = Blueprint("frontend", __name__)


@frontend.before_request
def login_check():
    request.user = None
    should_login = True
    if session.get("userid"):
        try:
            user = UserManager.get(session.get("userid"))
            request.user = user
            should_login = False
        except (ValueError, ObjectNotFound):
            session.clear()

    if should_login and request.endpoint != "frontend.login":
        flash("Please login")
        return redirect(
            url_for(
                "frontend.login",
                returnUrl=request.script_root
                + request.url[len(request.url_root) - 1 :],
            )
        )


@frontend.before_request
def scan_status():
    if not request.user or not request.user.admin:
        return

    try:
        scanned = DaemonClient(
            current_app.config["DAEMON"]["socket"]
        ).get_scanning_progress()
        if scanned is not None:
            flash("Scanning in progress, {} files scanned.".format(scanned))
    except DaemonUnavailableError:
        pass


@frontend.route("/")
def index():
    stats = {
        "artists": Artist.select().count(),
        "albums": Album.select().count(),
        "tracks": Track.select().count(),
    }
    return render_template("home.html", stats=stats)


def admin_only(f):
    @wraps(f)
    def decorated_func(*args, **kwargs):
        if not request.user or not request.user.admin:
            return redirect(url_for("frontend.index"))
        return f(*args, **kwargs)

    return decorated_func


from .user import *
from .folder import *
from .playlist import *



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from base64 import b64encode, b64decode
from os import urandom
from pony.orm import db_session, commit, ObjectNotFound

from supysonic.db import Meta


__key_cache = {}


def get_secret_key(keyname):
    if keyname in __key_cache:
        return __key_cache[keyname]

    with db_session():
        # Commit both at enter and exit. The metadb/db split (from supysonic.db)
        # confuses Pony which can either error or hang when this method is called
        commit()
        try:
            key = b64decode(Meta[keyname].value)
        except ObjectNotFound:
            key = urandom(128)
            Meta(key=keyname, value=b64encode(key).decode())
        commit()

    __key_cache[keyname] = key
    return key



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

API_VERSION = "1.10.2"

import binascii
import uuid
from flask import request
from flask import Blueprint
from pony.orm import ObjectNotFound, TransactionIntegrityError
from pony.orm import commit

from ..db import ClientPrefs, Folder
from ..managers.user import UserManager

from .exceptions import GenericError, Unauthorized
from .formatters import JSONFormatter, JSONPFormatter, XMLFormatter

api = Blueprint("api", __name__)


def api_routing(endpoint):
    def decorator(func):
        viewendpoint = "{}.view".format(endpoint)
        api.add_url_rule(endpoint, view_func=func, methods=["GET", "POST"])
        api.add_url_rule(viewendpoint, view_func=func, methods=["GET", "POST"])
        return func

    return decorator


@api.before_request
def set_formatter():
    """Return a function to create the response."""
    f, callback = map(request.values.get, ["f", "callback"])
    if f == "jsonp":
        request.formatter = JSONPFormatter(callback)
    elif f == "json":
        request.formatter = JSONFormatter()
    else:
        request.formatter = XMLFormatter()


def decode_password(password):
    if not password.startswith("enc:"):
        return password

    try:
        return binascii.unhexlify(password[4:].encode("utf-8")).decode("utf-8")
    except ValueError:
        return password


@api.before_request
def authorize():
    if request.authorization:
        user = UserManager.try_auth(
            request.authorization.username, request.authorization.password
        )
        if user is not None:
            request.user = user
            return
        raise Unauthorized()

    username = request.values["u"]
    password = request.values["p"]
    password = decode_password(password)

    user = UserManager.try_auth(username, password)
    if user is None:
        raise Unauthorized()

    request.user = user


@api.before_request
def get_client_prefs():
    client = request.values["c"]
    try:
        request.client = ClientPrefs[request.user, client]
    except ObjectNotFound:
        try:
            request.client = ClientPrefs(user=request.user, client_name=client)
            commit()
        except TransactionIntegrityError:
            # We might have hit a race condition here, another request already created
            # the ClientPrefs. Issue #220
            # Reload the user or Pony will complain about different transactions
            request.user = UserManager.get(request.user.id)
            request.client = ClientPrefs[request.user, client]


def get_entity(cls, param="id"):
    eid = request.values[param]
    if cls == Folder:
        eid = int(eid)
    else:
        eid = uuid.UUID(eid)
    entity = cls[eid]
    return entity


def get_entity_id(cls, eid):
    """Return the entity ID as its proper type."""
    if cls == Folder:
        if isinstance(eid, uuid.UUID):
            raise GenericError("Invalid ID")
        try:
            return int(eid)
        except ValueError:
            raise GenericError("Invalid ID")
    try:
        return uuid.UUID(eid)
    except (AttributeError, ValueError):
        raise GenericError("Invalid ID")


from .errors import *

from .system import *
from .browse import *
from .user import *
from .albums_songs import *
from .media import *
from .annotation import *
from .chat import *
from .search import *
from .playlists import *
from .jukebox import *
from .radio import *
from .unsupported import *
from .scan import *



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2021 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import importlib
import os
import os.path

from click import command, option, Option
from click.exceptions import UsageError, ClickException
from click.types import Choice

from ..web import create_application

_servers = [
    e.name[:-3]
    for e in os.scandir(os.path.dirname(__file__))
    if not e.name.startswith("_") and e.name.endswith(".py")
]


class MutuallyExclusiveOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs[
                "help"
            ] = "{}  NOTE: This argument is mutually exclusive with arguments: [{}].".format(
                help, ex_str
            )
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with arguments `{}`.".format(
                    self.name, ", ".join(self.mutually_exclusive)
                )
            )

        return super().handle_parse_result(ctx, opts, args)


def get_server(name):
    return importlib.import_module("." + name, __package__).server


def find_first_available_server():
    for module in _servers:
        try:
            return get_server(module)
        except ImportError:
            pass

    return None


@command()
@option(
    "-S",
    "--server",
    type=Choice(_servers),
    help="Specify which WSGI server to use. Pick the first available if none is set",
)
@option(
    "-h",
    "--host",
    default="0.0.0.0",
    show_default=True,
    help="Hostname or IP address on which to listen",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("socket",),
)
@option(
    "-p",
    "--port",
    default=5722,
    show_default=True,
    help="TCP port on which to listen",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("socket",),
)
@option(
    "-s",
    "--socket",
    help="Unix socket on which to bind to, Can't be used with --host and --port",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=("host", "port"),
)
@option(
    "--processes",
    type=int,
    help="Number of processes to spawn. May not be supported by all servers",
)
@option(
    "--threads",
    type=int,
    help="Number of threads used to process application logic. May not be supported by all servers",
)
def main(server, host, port, socket, processes, threads):
    """Starts the Supysonic web server"""

    if server is None:
        server = find_first_available_server()
        if server is None:
            raise ClickException(
                "Couldn't load any server, please install one of {}".format(_servers)
            )
    else:
        try:
            server = get_server(server)
        except ImportError:
            raise ClickException(
                "Couldn't load {}, please install it first".format(server)
            )

    if socket is not None:
        host = None
        port = None

    app = create_application()
    server(
        app, host=host, port=port, socket=socket, processes=processes, threads=threads
    ).run()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import time

from multiprocessing.connection import Listener, Client
from pony.orm import db_session, select
from threading import Thread, Event

from .client import DaemonCommand
from ..db import Folder
from ..jukebox import Jukebox
from ..scanner import Scanner
from ..utils import get_secret_key
from ..watcher import SupysonicWatcher

__all__ = ["Daemon"]

logger = logging.getLogger(__name__)


class Daemon:
    def __init__(self, config):
        self.__config = config
        self.__listener = None
        self.__watcher = None
        self.__scanner = None
        self.__jukebox = None
        self.__stopped = Event()

    watcher = property(lambda self: self.__watcher)
    scanner = property(lambda self: self.__scanner)
    jukebox = property(lambda self: self.__jukebox)

    def __handle_connection(self, connection):
        cmd = connection.recv()
        logger.debug("Received %s", cmd)
        if cmd is None:
            pass
        elif isinstance(cmd, DaemonCommand):
            cmd.apply(connection, self)
        else:
            logger.warn("Received unknown command %s", cmd)

    def run(self):
        self.__listener = Listener(
            address=self.__config.DAEMON["socket"], authkey=get_secret_key("daemon_key")
        )
        logger.info("Listening to %s", self.__listener.address)

        if self.__config.DAEMON["run_watcher"]:
            self.__watcher = SupysonicWatcher(self.__config)
            self.__watcher.start()

        if self.__config.DAEMON["jukebox_command"]:
            self.__jukebox = Jukebox(self.__config.DAEMON["jukebox_command"])

        Thread(target=self.__listen).start()
        while not self.__stopped.is_set():
            time.sleep(1)

    def __listen(self):
        while not self.__stopped.is_set():
            conn = self.__listener.accept()
            self.__handle_connection(conn)

        self.__listener.close()

    def start_scan(self, folders=[], force=False):
        if not folders:
            with db_session:
                folders = select(f.name for f in Folder if f.root)[:]

        if self.__scanner is not None and self.__scanner.is_alive():
            for f in folders:
                self.__scanner.queue_folder(f)
            return

        extensions = self.__config.BASE["scanner_extensions"]
        if extensions:
            extensions = extensions.split(" ")

        self.__scanner = Scanner(
            force=force,
            extensions=extensions,
            follow_symlinks=self.__config.BASE["follow_symlinks"],
            on_folder_start=self.__unwatch,
            on_folder_end=self.__watch,
        )
        for f in folders:
            self.__scanner.queue_folder(f)

        self.__scanner.start()

    def __watch(self, folder):
        if self.__watcher is not None:
            self.__watcher.add_folder(folder.path)

    def __unwatch(self, folder):
        if self.__watcher is not None:
            self.__watcher.remove_folder(folder.path)

    def terminate(self):
        self.__stopped.set()
        with Client(self.__listener.address, authkey=self.__listener._authkey) as c:
            c.send(None)

        if self.__scanner is not None:
            self.__scanner.stop()
            self.__scanner.join()
        if self.__watcher is not None:
            self.__watcher.stop()
        if self.__jukebox is not None:
            self.__jukebox.terminate()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#               2018-2019 Carey 'pR0Ps' Metcalfe
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import mimetypes

from flask import Flask
from os import makedirs, path
from pony.orm import db_session

from .config import IniConfig
from .cache import Cache
from .db import init_database
from .utils import get_secret_key

logger = logging.getLogger(__package__)


def create_application(config=None):
    global app

    # Flask!
    app = Flask(__name__)
    app.config.from_object("supysonic.config.DefaultConfig")

    if not config:  # pragma: nocover
        config = IniConfig.from_common_locations()
    app.config.from_object(config)

    # Set loglevel
    logfile = app.config["WEBAPP"]["log_file"]
    if logfile:  # pragma: nocover
        from logging.handlers import TimedRotatingFileHandler

        handler = TimedRotatingFileHandler(logfile, when="midnight")
        handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
        )
        logger.addHandler(handler)
    loglevel = app.config["WEBAPP"]["log_level"]
    if loglevel:
        logger.setLevel(getattr(logging, loglevel.upper(), logging.NOTSET))

    # Initialize database
    init_database(app.config["BASE"]["database_uri"])
    app.wsgi_app = db_session(app.wsgi_app)

    # Insert unknown mimetypes
    for k, v in app.config["MIMETYPES"].items():
        extension = "." + k.lower()
        if extension not in mimetypes.types_map:
            mimetypes.add_type(v, extension, False)

    # Initialize Cache objects
    # Max size is MB in the config file but Cache expects bytes
    cache_dir = app.config["WEBAPP"]["cache_dir"]
    max_size_cache = app.config["WEBAPP"]["cache_size"] * 1024 ** 2
    max_size_transcodes = app.config["WEBAPP"]["transcode_cache_size"] * 1024 ** 2
    app.cache = Cache(path.join(cache_dir, "cache"), max_size_cache)
    app.transcode_cache = Cache(path.join(cache_dir, "transcodes"), max_size_transcodes)

    # Test for the cache directory
    cache_path = app.config["WEBAPP"]["cache_dir"]
    if not path.exists(cache_path):
        makedirs(cache_path)  # pragma: nocover

    # Read or create secret key
    app.secret_key = get_secret_key("cookies_secret")

    # Import app sections
    if app.config["WEBAPP"]["mount_webui"]:
        from .frontend import frontend

        app.register_blueprint(frontend)
    if app.config["WEBAPP"]["mount_api"]:
        from .api import api

        app.register_blueprint(api, url_prefix="/rest")

    return app



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import uuid

from flask import Response, flash, redirect, render_template, request, url_for
from pony.orm import ObjectNotFound

from ..db import Playlist

from . import frontend


@frontend.route("/playlist")
def playlist_index():
    return render_template(
        "playlists.html",
        mine=Playlist.select(lambda p: p.user == request.user),
        others=Playlist.select(lambda p: p.user != request.user and p.public),
    )


@frontend.route("/playlist/<uid>")
def playlist_details(uid):
    try:
        uid = uuid.UUID(uid)
    except ValueError:
        flash("Invalid playlist id")
        return redirect(url_for("frontend.playlist_index"))

    try:
        playlist = Playlist[uid]
    except ObjectNotFound:
        flash("Unknown playlist")
        return redirect(url_for("frontend.playlist_index"))

    return render_template("playlist.html", playlist=playlist)


@frontend.route("/playlist/<uid>/export")
def playlist_export(uid):
    try:
        uid = uuid.UUID(uid)
    except ValueError:
        flash("Invalid playlist id")
        return redirect(url_for("frontend.playlist_index"))

    try:
        playlist = Playlist[uid]
    except ObjectNotFound:
        flash("Unknown playlist")
        return redirect(url_for("frontend.playlist_index"))

    return Response(
        render_template("playlist_export.m3u", playlist=playlist),
        mimetype="audio/mpegurl",
        headers={
            "Content-disposition": "attachment; filename={}.m3u".format(playlist.name)
        },
    )


@frontend.route("/playlist/<uid>", methods=["POST"])
def playlist_update(uid):
    try:
        uid = uuid.UUID(uid)
    except ValueError:
        flash("Invalid playlist id")
        return redirect(url_for("frontend.playlist_index"))

    try:
        playlist = Playlist[uid]
    except ObjectNotFound:
        flash("Unknown playlist")
        return redirect(url_for("frontend.playlist_index"))

    if playlist.user.id != request.user.id:
        flash("You're not allowed to edit this playlist")
    elif not request.form.get("name"):
        flash("Missing playlist name")
    else:
        playlist.name = request.form.get("name")
        playlist.public = request.form.get("public") in (
            True,
            "True",
            1,
            "1",
            "on",
            "checked",
        )
        flash("Playlist updated.")

    return playlist_details(str(uid))


@frontend.route("/playlist/del/<uid>")
def playlist_delete(uid):
    try:
        uid = uuid.UUID(uid)
    except ValueError:
        flash("Invalid playlist id")
        return redirect(url_for("frontend.playlist_index"))

    try:
        playlist = Playlist[uid]
    except ObjectNotFound:
        flash("Unknown playlist")
        return redirect(url_for("frontend.playlist_index"))

    if playlist.user.id != request.user.id:
        flash("You're not allowed to delete this playlist")
    else:
        playlist.delete()
        flash("Playlist deleted")

    return redirect(url_for("frontend.playlist_index"))



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import re
import string

from flask import current_app, request
from pony.orm import ObjectNotFound, select, count

from ..db import Folder, Artist, Album, Track

from . import get_entity, get_entity_id, api_routing


@api_routing("/getMusicFolders")
def list_folders():
    return request.formatter(
        "musicFolders",
        dict(
            musicFolder=[
                dict(id=str(f.id), name=f.name)
                for f in Folder.select(lambda f: f.root).order_by(Folder.name)
            ]
        ),
    )


def build_ignored_articles_pattern():
    articles = current_app.config["WEBAPP"]["index_ignored_prefixes"]
    if articles is None:
        return None

    articles = articles.split()
    if not articles:
        return None

    return r"^(" + r" |".join(re.escape(a) for a in articles) + r" )"


def ignored_articles_str():
    articles = current_app.config["WEBAPP"]["index_ignored_prefixes"]
    if articles is None:
        return ""

    return " ".join(articles.split())


@api_routing("/getIndexes")
def list_indexes():
    musicFolderId = request.values.get("musicFolderId")
    ifModifiedSince = request.values.get("ifModifiedSince")
    if ifModifiedSince:
        ifModifiedSince = int(ifModifiedSince) / 1000

    if musicFolderId is None:
        folders = Folder.select(lambda f: f.root)[:]
    else:
        mfid = get_entity_id(Folder, musicFolderId)
        folder = Folder[mfid]
        if not folder.root:
            raise ObjectNotFound(Folder, mfid)

        folders = [folder]

    last_modif = max(map(lambda f: f.last_scan, folders))
    if ifModifiedSince is not None and last_modif < ifModifiedSince:
        return request.formatter(
            "indexes",
            dict(
                lastModified=last_modif * 1000, ignoredArticles=ignored_articles_str()
            ),
        )

    # The XSD lies, we don't return artists but a directory structure
    artists = []
    children = []
    for f in folders:
        artists += f.children.select()[:]
        children += f.tracks.select()[:]

    indexes = dict()
    pattern = build_ignored_articles_pattern()
    for artist in artists:
        name = artist.name
        if pattern:
            name = re.sub(pattern, "", name, flags=re.I)
        index = name[0].upper()
        if index in string.digits:
            index = "#"
        elif index not in string.ascii_letters:
            index = "?"

        if index not in indexes:
            indexes[index] = []

        indexes[index].append((artist, name))

    return request.formatter(
        "indexes",
        dict(
            lastModified=last_modif * 1000,
            ignoredArticles=ignored_articles_str(),
            index=[
                dict(
                    name=k,
                    artist=[
                        a.as_subsonic_artist(request.user)
                        for a, _ in sorted(v, key=lambda t: t[1].lower())
                    ],
                )
                for k, v in sorted(indexes.items())
            ],
            child=[
                c.as_subsonic_child(request.user, request.client)
                for c in sorted(children, key=lambda t: t.sort_key())
            ],
        ),
    )


@api_routing("/getMusicDirectory")
def show_directory():
    res = get_entity(Folder)
    return request.formatter(
        "directory", res.as_subsonic_directory(request.user, request.client)
    )


@api_routing("/getGenres")
def list_genres():
    return request.formatter(
        "genres",
        dict(
            genre=[
                dict(value=genre, songCount=sc, albumCount=ac)
                for genre, sc, ac in select(
                    (t.genre, count(), count(t.album)) for t in Track if t.genre
                )
            ]
        ),
    )


@api_routing("/getArtists")
def list_artists():
    # According to the API page, there are no parameters?
    indexes = dict()
    pattern = build_ignored_articles_pattern()
    for artist in Artist.select():
        name = artist.name or "?"
        if pattern:
            name = re.sub(pattern, "", name, flags=re.I)
        index = name[0].upper()
        if index in string.digits:
            index = "#"
        elif index not in string.ascii_letters:
            index = "?"

        if index not in indexes:
            indexes[index] = []

        indexes[index].append((artist, name))

    return request.formatter(
        "artists",
        dict(
            ignoredArticles=ignored_articles_str(),
            index=[
                dict(
                    name=k,
                    artist=[
                        a.as_subsonic_artist(request.user)
                        for a, _ in sorted(v, key=lambda t: t[1].lower())
                    ],
                )
                for k, v in sorted(indexes.items())
            ],
        ),
    )


@api_routing("/getArtist")
def artist_info():
    res = get_entity(Artist)
    info = res.as_subsonic_artist(request.user)
    albums = set(res.albums)
    albums |= {t.album for t in res.tracks}
    info["album"] = [
        a.as_subsonic_album(request.user)
        for a in sorted(albums, key=lambda a: a.sort_key())
    ]

    return request.formatter("artist", info)


@api_routing("/getAlbum")
def album_info():
    res = get_entity(Album)
    info = res.as_subsonic_album(request.user)
    info["song"] = [
        t.as_subsonic_child(request.user, request.client)
        for t in sorted(res.tracks, key=lambda t: t.sort_key())
    ]

    return request.formatter("album", info)


@api_routing("/getSong")
def track_info():
    res = get_entity(Track)
    return request.formatter(
        "song", res.as_subsonic_child(request.user, request.client)
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from datetime import timedelta
from flask import request
from pony.orm import select, desc, avg, max, min, count, between, distinct

from ..db import (
    Folder,
    Album,
    Track,
    StarredFolder,
    StarredArtist,
    StarredAlbum,
    StarredTrack,
    User,
)
from ..db import now

from . import api_routing
from .exceptions import GenericError, NotFound


@api_routing("/getRandomSongs")
def rand_songs():
    size = request.values.get("size", "10")
    genre, fromYear, toYear, musicFolderId = map(
        request.values.get, ["genre", "fromYear", "toYear", "musicFolderId"]
    )

    size = int(size) if size else 10
    fromYear = int(fromYear) if fromYear else None
    toYear = int(toYear) if toYear else None
    fid = None
    if musicFolderId:
        try:
            fid = int(musicFolderId)
        except ValueError:
            raise ValueError("Invalid folder ID")

    query = Track.select()
    if fromYear:
        query = query.filter(lambda t: t.year >= fromYear)
    if toYear:
        query = query.filter(lambda t: t.year <= toYear)
    if genre:
        query = query.filter(lambda t: t.genre == genre)
    if fid:
        if not Folder.exists(id=fid, root=True):
            raise NotFound("Folder")

        query = query.filter(lambda t: t.root_folder.id == fid)

    return request.formatter(
        "randomSongs",
        dict(
            song=[
                t.as_subsonic_child(request.user, request.client)
                for t in query.without_distinct().random(size)
            ]
        ),
    )


@api_routing("/getAlbumList")
def album_list():
    ltype = request.values["type"]

    size, offset = map(request.values.get, ["size", "offset"])
    size = int(size) if size else 10
    offset = int(offset) if offset else 0

    query = select(t.folder for t in Track)
    if ltype == "random":
        return request.formatter(
            "albumList",
            dict(
                album=[
                    a.as_subsonic_child(request.user)
                    for a in distinct(query.random(size))
                ]
            ),
        )
    elif ltype == "newest":
        query = query.sort_by(desc(Folder.created)).distinct()
    elif ltype == "highest":
        query = query.sort_by(lambda f: desc(avg(f.ratings.rating)))
    elif ltype == "frequent":
        query = query.sort_by(lambda f: desc(avg(f.tracks.play_count)))
    elif ltype == "recent":
        query = select(
            t.folder for t in Track if max(t.folder.tracks.last_play) is not None
        ).sort_by(lambda f: desc(max(f.tracks.last_play)))
    elif ltype == "starred":
        query = select(
            s.starred
            for s in StarredFolder
            if s.user.id == request.user.id and count(s.starred.tracks) > 0
        )
    elif ltype == "alphabeticalByName":
        query = query.sort_by(Folder.name).distinct()
    elif ltype == "alphabeticalByArtist":
        query = query.sort_by(lambda f: f.parent.name + f.name)
    elif ltype == "byYear":
        startyear = int(request.values["fromYear"])
        endyear = int(request.values["toYear"])
        query = query.where(
            lambda t: between(t.year, min(startyear, endyear), max(startyear, endyear))
        )
        if endyear < startyear:
            query = query.sort_by(lambda f: desc(min(f.tracks.year)))
        else:
            query = query.sort_by(lambda f: min(f.tracks.year))
    elif ltype == "byGenre":
        genre = request.values["genre"]
        query = query.where(lambda t: t.genre == genre)
    else:
        raise GenericError("Unknown search type")

    return request.formatter(
        "albumList",
        dict(
            album=[f.as_subsonic_child(request.user) for f in query.limit(size, offset)]
        ),
    )


@api_routing("/getAlbumList2")
def album_list_id3():
    ltype = request.values["type"]

    size, offset = map(request.values.get, ["size", "offset"])
    size = int(size) if size else 10
    offset = int(offset) if offset else 0

    query = Album.select()
    if ltype == "random":
        return request.formatter(
            "albumList2",
            dict(album=[a.as_subsonic_album(request.user) for a in query.random(size)]),
        )
    elif ltype == "newest":
        query = query.order_by(lambda a: desc(min(a.tracks.created)))
    elif ltype == "frequent":
        query = query.order_by(lambda a: desc(avg(a.tracks.play_count)))
    elif ltype == "recent":
        query = Album.select(lambda a: max(a.tracks.last_play) is not None).order_by(
            lambda a: desc(max(a.tracks.last_play))
        )
    elif ltype == "starred":
        query = select(s.starred for s in StarredAlbum if s.user.id == request.user.id)
    elif ltype == "alphabeticalByName":
        query = query.order_by(Album.name)
    elif ltype == "alphabeticalByArtist":
        query = query.order_by(lambda a: a.artist.name + a.name)
    elif ltype == "byYear":
        startyear = int(request.values["fromYear"])
        endyear = int(request.values["toYear"])
        query = query.where(
            lambda a: between(
                min(a.tracks.year), min(startyear, endyear), max(startyear, endyear)
            )
        )
        if endyear < startyear:
            query = query.order_by(lambda a: desc(min(a.tracks.year)))
        else:
            query = query.order_by(lambda a: min(a.tracks.year))
    elif ltype == "byGenre":
        genre = request.values["genre"]
        query = query.where(lambda a: genre in a.tracks.genre)
    else:
        raise GenericError("Unknown search type")

    return request.formatter(
        "albumList2",
        dict(
            album=[f.as_subsonic_album(request.user) for f in query.limit(size, offset)]
        ),
    )


@api_routing("/getSongsByGenre")
def songs_by_genre():
    genre = request.values["genre"]

    count, offset = map(request.values.get, ["count", "offset"])
    count = int(count) if count else 10
    offset = int(offset) if offset else 0

    query = select(t for t in Track if t.genre == genre).limit(count, offset)
    return request.formatter(
        "songsByGenre",
        dict(song=[t.as_subsonic_child(request.user, request.client) for t in query]),
    )


@api_routing("/getNowPlaying")
def now_playing():
    query = User.select(
        lambda u: u.last_play is not None
        and u.last_play_date + timedelta(minutes=3) > now()
    )

    return request.formatter(
        "nowPlaying",
        dict(
            entry=[
                dict(
                    u.last_play.as_subsonic_child(request.user, request.client),
                    username=u.name,
                    minutesAgo=(now() - u.last_play_date).seconds / 60,
                    playerId=0,
                )
                for u in query
            ]
        ),
    )


@api_routing("/getStarred")
def get_starred():
    folders = select(s.starred for s in StarredFolder if s.user.id == request.user.id)

    return request.formatter(
        "starred",
        dict(
            artist=[
                sf.as_subsonic_artist(request.user)
                for sf in folders.filter(lambda f: count(f.tracks) == 0)
            ],
            album=[
                sf.as_subsonic_child(request.user)
                for sf in folders.filter(lambda f: count(f.tracks) > 0)
            ],
            song=[
                st.as_subsonic_child(request.user, request.client)
                for st in select(
                    s.starred for s in StarredTrack if s.user.id == request.user.id
                )
            ],
        ),
    )


@api_routing("/getStarred2")
def get_starred_id3():
    return request.formatter(
        "starred2",
        dict(
            artist=[
                sa.as_subsonic_artist(request.user)
                for sa in select(
                    s.starred for s in StarredArtist if s.user.id == request.user.id
                )
            ],
            album=[
                sa.as_subsonic_album(request.user)
                for sa in select(
                    s.starred for s in StarredAlbum if s.user.id == request.user.id
                )
            ],
            song=[
                st.as_subsonic_child(request.user, request.client)
                for st in select(
                    s.starred for s in StarredTrack if s.user.id == request.user.id
                )
            ],
        ),
    )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import time

from flask import current_app, request
from pony.orm import delete
from pony.orm import ObjectNotFound

from ..db import Track, Album, Artist, Folder
from ..db import StarredTrack, StarredAlbum, StarredArtist, StarredFolder
from ..db import RatingTrack, RatingFolder
from ..lastfm import LastFm

from . import get_entity, get_entity_id, api_routing
from .exceptions import AggregateException, GenericError, MissingParameter, NotFound


def star_single(cls, starcls, eid):
    """Stars an entity

    :param cls: entity class, Folder, Artist, Album or Track
    :param starcls: matching starred class, StarredFolder, StarredArtist, StarredAlbum or StarredTrack
    :param eid: id of the entity to star
    """

    try:
        e = cls[eid]
    except ObjectNotFound:
        raise NotFound("{} {}".format(cls.__name__, eid))

    try:
        starcls[request.user, eid]
        raise GenericError("{} {} already starred".format(cls.__name__, eid))
    except ObjectNotFound:
        pass

    starcls(user=request.user, starred=e)


def unstar_single(cls, starcls, eid):
    """Unstars an entity

    :param cls: entity class, Folder, Artist, Album or Track
    :param starcls: matching starred class, StarredFolder, StarredArtist, StarredAlbum or StarredTrack
    :param eid: id of the entity to unstar
    """

    delete(s for s in starcls if s.user.id == request.user.id and s.starred.id == eid)


def handle_star_request(func):
    id, albumId, artistId = map(request.values.getlist, ["id", "albumId", "artistId"])

    if not id and not albumId and not artistId:
        raise MissingParameter("id, albumId or artistId")

    errors = []
    for eid in id:
        try:
            tid = get_entity_id(Track, eid)
        except GenericError:
            tid = None
        try:
            fid = get_entity_id(Folder, eid)
        except GenericError:
            fid = None
        err = None

        if tid is None and fid is None:
            raise GenericError("Invalid ID")

        if tid is not None:
            try:
                func(Track, StarredTrack, tid)
            except Exception as e:
                err = e
        else:
            try:
                func(Folder, StarredFolder, fid)
            except Exception as e:
                err = e

        if err:
            errors.append(err)

    for alId in albumId:
        alb_id = get_entity_id(Album, alId)
        try:
            func(Album, StarredAlbum, alb_id)
        except Exception as e:
            errors.append(e)

    for arId in artistId:
        art_id = get_entity_id(Artist, arId)
        try:
            func(Artist, StarredArtist, art_id)
        except Exception as e:
            errors.append(e)

    if errors:
        raise AggregateException(errors)
    return request.formatter.empty


@api_routing("/star")
def star():
    return handle_star_request(star_single)


@api_routing("/unstar")
def unstar():
    return handle_star_request(unstar_single)


@api_routing("/setRating")
def rate():
    id = request.values["id"]
    rating = request.values["rating"]

    try:
        tid = get_entity_id(Track, id)
    except GenericError:
        tid = None
    try:
        fid = get_entity_id(Folder, id)
    except GenericError:
        fid = None
    uid = None
    rating = int(rating)

    if tid is None and fid is None:
        raise GenericError("Invalid ID")

    if not 0 <= rating <= 5:
        raise GenericError("rating must be between 0 and 5 (inclusive)")

    if rating == 0:
        if tid is not None:
            delete(
                r
                for r in RatingTrack
                if r.user.id == request.user.id and r.rated.id == tid
            )
        else:
            delete(
                r
                for r in RatingFolder
                if r.user.id == request.user.id and r.rated.id == fid
            )
    else:
        if tid is not None:
            rated = Track[tid]
            rating_cls = RatingTrack
            uid = tid
        else:
            rated = Folder[fid]
            rating_cls = RatingFolder
            uid = fid

        try:
            rating_info = rating_cls[request.user, uid]
            rating_info.rating = rating
        except ObjectNotFound:
            rating_cls(user=request.user, rated=rated, rating=rating)

    return request.formatter.empty


@api_routing("/scrobble")
def scrobble():
    res = get_entity(Track)
    t, submission = map(request.values.get, ["time", "submission"])
    t = int(t) / 1000 if t else int(time.time())

    lfm = LastFm(current_app.config["LASTFM"], request.user)

    if submission in (None, "", True, "true", "True", 1, "1"):
        lfm.scrobble(res, t)
    else:
        lfm.now_playing(res)

    return request.formatter.empty
